
#include "DetectorConstruction.hh"
#include "DetectorConstructionMessenger.hh"

#include "G4Box.hh"
#include "G4Tubs.hh"
#include "G4LogicalVolume.hh"
#include "G4PVPlacement.hh"
#include "G4PVReplica.hh"

#include "G4RunManager.hh"

#include "G4GeometryManager.hh"
#include "G4PhysicalVolumeStore.hh"
#include "G4LogicalVolumeStore.hh"
#include "G4SolidStore.hh"

#include "G4VisAttributes.hh"
#include "G4Colour.hh"

#include "G4UnitsTable.hh"
#include "G4ios.hh"
#include "G4ThreeVector.hh"

#include "G4NistManager.hh"

#include "ScintillatorSD.hh"
#include "G4SDManager.hh"
#include "G4VSensitiveDetector.hh"

#include "globals.hh"

#include "TSimulationParameterNeutronDetector.hh"

#include "TVector3.h"

#include <sstream>
#include <algorithm>

DetectorConstruction::DetectorConstruction()
  : fNeutNumber(1), fNeutSize(10*cm, 10*cm, 10*cm), fNeutGap(0), fNeutRotationMode(0),
    fVetoNumber(0), fVetoSize(0, 0, 0),fVetoGap(0, 0, 0), fVetoRotationMode(0),
    fNeutLayerNumber(1), fNeutLayerGap(0),
    fNeutVetoGap(0),
    fStuckNumber(1), fStuckGap(0),
    fPosition(0, 0, 10.*m),
    fLogicNeut(0), fLogicNeutLayer(0), fLogicStuck(0),
    fLogicVeto(0), fLogicVetoLayer(0), fLogicWorld(0),
    fScintillatorSD(0)
{
  fDetectorConstructionMessenger = new DetectorConstructionMessenger(this);

  fNeutMaterial = G4NistManager::Instance()->FindOrBuildMaterial("G4_PLASTIC_SC_VINYLTOLUENE");
  fNeutLayerMaterial = G4NistManager::Instance()->FindOrBuildMaterial("G4_AIR");
  fStuckMaterial = G4NistManager::Instance()->FindOrBuildMaterial("G4_AIR");
  fVetoMaterial = G4NistManager::Instance()->FindOrBuildMaterial("G4_PLASTIC_SC_VINYLTOLUENE");
  fVetoLayerMaterial = G4NistManager::Instance()->FindOrBuildMaterial("G4_AIR");
  fWorldMaterial = G4NistManager::Instance()->FindOrBuildMaterial("G4_AIR");
}

DetectorConstruction::~DetectorConstruction()
{
  delete fDetectorConstructionMessenger;
}

G4VPhysicalVolume* DetectorConstruction::Construct()
{
  // Cleanup old geometry
  G4GeometryManager::GetInstance()->OpenGeometry();
  G4PhysicalVolumeStore::GetInstance()->Clean();
  G4LogicalVolumeStore::GetInstance()->Clean();
  G4SolidStore::GetInstance()->Clean();

  // if not using Veto
  if(fVetoNumber == 0){
    fVetoSize.set(0, 0, 0);
    fVetoGap.set(0, 0, 0);
    fNeutVetoGap = 0;
  }

  // Neut LV
  G4ThreeVector neutPos(0, -(fNeutSize.y()+fNeutGap)*(fNeutNumber-1)/2, 0);
  G4ThreeVector neutPosShift(0, fNeutSize.y()+fNeutGap, 0);
  G4Box* solidNeut = new G4Box("Neut",
			       fNeutSize.x()/2,
			       fNeutSize.y()/2,
			       fNeutSize.z()/2);
  fLogicNeut = new G4LogicalVolume(solidNeut,fNeutMaterial,"Neut");

  // Veto LV
  G4ThreeVector vetoPos(0, -(fVetoSize.y()+fVetoGap.y())*(fVetoNumber-1)/2, fVetoGap.z()/2);
  G4ThreeVector vetoPosShift(0, fVetoSize.y()+fVetoGap.y(), fVetoGap.z()); // fVetoGap.z()/2 *= -1
  G4Box* solidVeto = 0;
  if(fVetoNumber != 0){ 
    solidVeto = new G4Box("Veto",
			  fVetoSize.x()/2,
			  fVetoSize.y()/2,
			  fVetoSize.z()/2);
    fLogicVeto = new G4LogicalVolume(solidVeto,fVetoMaterial,"Veto");
  }

  // Neut and Veto Layer LV
  G4ThreeVector neutLayerSize(fNeutSize.x(),
			      (fNeutSize.y()+fNeutGap)*fNeutNumber-fNeutGap,
			      fNeutSize.z());
  G4ThreeVector vetoLayerSize(fVetoSize.x(),
			      (fVetoSize.y()+fVetoGap.y())*fVetoNumber-fVetoGap.y(),
			      fVetoSize.z()+std::abs(fVetoGap.z()));
  G4double neutStuckSizeZ = (neutLayerSize.z()+fNeutLayerGap)*fNeutLayerNumber-fNeutLayerGap;
  G4double vetoStuckSizeZ = vetoLayerSize.z()+fNeutVetoGap;
  G4ThreeVector neutLayerPos(0, 0, -(neutStuckSizeZ-vetoStuckSizeZ-fNeutSize.z())/2);
  G4ThreeVector vetoLayerPos(0, 0, -neutStuckSizeZ/2);
  G4ThreeVector neutLayerPosShift(0, 0, neutLayerSize.z()+fNeutLayerGap);
  G4Box* solidNeutLayer = new G4Box("NeutLayer",
				    neutLayerSize.x()/2,
				    neutLayerSize.y()/2,
				    neutLayerSize.z()/2);
  fLogicNeutLayer = new G4LogicalVolume(solidNeutLayer,fNeutLayerMaterial,"NeutLayer");
  G4Box* solidVetoLayer = 0;
  if(fVetoNumber != 0){ 
    solidVetoLayer = new G4Box("VetoLayer",
			       vetoLayerSize.x()/2,
			       vetoLayerSize.y()/2,
			       vetoLayerSize.z()/2);
    fLogicVetoLayer = new G4LogicalVolume(solidVetoLayer,fVetoLayerMaterial,"VetoLayer");
  }

  // Neutron Detector Stuck LV
  G4double neutLongSide = std::max(neutLayerSize.x(), neutLayerSize.y());
  G4double vetoLongSide = std::max(vetoLayerSize.x(), vetoLayerSize.y());
  G4ThreeVector stuckSize(std::max(neutLongSide, vetoLongSide),
			  std::max(neutLongSide, vetoLongSide),
			  neutStuckSizeZ+vetoStuckSizeZ);
  G4ThreeVector stuckPos(fPosition.x()+0.1*mm,
			 fPosition.y()+0.1*mm,
			 fPosition.z()+(neutStuckSizeZ-vetoStuckSizeZ)/2); // neutの面
  G4ThreeVector stuckPosShift(0, 0, stuckSize.z()+fStuckGap);

  G4Box* solidStuck = new G4Box("Stuck",
				stuckSize.x()/2,
				stuckSize.y()/2,
				stuckSize.z()/2);
  fLogicStuck = new G4LogicalVolume(solidStuck,fStuckMaterial,"Stuck");

  // World LV
  G4ThreeVector worldSize(1.2*2*(std::abs(stuckPos.x())+stuckSize.x()/2),
			  1.2*2*(std::abs(stuckPos.y())+stuckSize.y()/2),
			  1.2*2*(std::abs(stuckPos.z())+stuckSize.z()/2
				 +(stuckSize.z()+fStuckGap)*(fStuckNumber-1)));
  G4Box* solidWorld = new G4Box("World",
			    worldSize.x()/2,
			    worldSize.y()/2,
			    worldSize.z()/2);
  fLogicWorld = new G4LogicalVolume(solidWorld,fWorldMaterial,"World");
  
  // Neut PV
  G4String neutName("Neut");
  if(fNeutGap == 0){
    new G4PVReplica(neutName,
		    fLogicNeut,
		    fLogicNeutLayer,
		    kYAxis,
		    fNeutNumber,
		    fNeutSize.y());
  }else{
    for(G4int i = 0; i < fNeutNumber; ++i){
      new G4PVPlacement(0,
			neutPos,
			fLogicNeut,
			neutName,
			fLogicNeutLayer,
			false,
			i);
      neutPos += neutPosShift;
    }
  }    

  // Veto PV
  if(fVetoNumber != 0){
    G4String vetoName("Veto");
    if(fVetoGap.y() == 0 && fVetoGap.z() == 0){
      new G4PVReplica(vetoName,
		      fLogicVeto,
		      fLogicVetoLayer,
		      kYAxis,
		      fVetoNumber,
		      fVetoSize.y());
    }else{
      for(G4int i = 0; i < fVetoNumber; ++i){
	new G4PVPlacement(0,
			  vetoPos,
			  fLogicVeto,
			  vetoName,
			  fLogicVetoLayer,
			  false,
			  i);
	vetoPosShift.setZ(-1*vetoPosShift.z());
	vetoPos += vetoPosShift;
      }
    }    
  }

  // Neut Mother PV
  G4String neutLayerName("NeutLayer");
  for(G4int i = 0; i < fNeutLayerNumber; ++i){
    G4RotationMatrix armRot= G4RotationMatrix();
    if(fNeutRotationMode == 1){ 
      armRot.rotateZ(-90*deg);
    }else if(fNeutRotationMode == 2){
      armRot.rotateZ(-90*deg*(i%2));
    }
    new G4PVPlacement(G4Transform3D(armRot, neutLayerPos),
		      fLogicNeutLayer,
		      neutLayerName,
		      fLogicStuck,
		      false,
		      i);
    neutLayerPos += neutLayerPosShift;
  }

  // Veto Mother PV
  if(fVetoNumber != 0){
    G4String vetoLayerName("VetoLayer");
    {
      G4int i = 0;
      G4RotationMatrix armRot= G4RotationMatrix();
      if(fVetoRotationMode == 1){ 
	armRot.rotateZ(-90*deg);
      }
      new G4PVPlacement(G4Transform3D(armRot, vetoLayerPos),
			fLogicVetoLayer,
			vetoLayerName,
			fLogicStuck,
			false,
			i);
    }
  }

  // Module Grand-Mother PV
  G4String stuckName("Stuck");
  for(G4int i = 0; i < fStuckNumber; ++i){
    new G4PVPlacement(0,
		      stuckPos,
		      fLogicStuck,
		      stuckName,
		      fLogicWorld,
		      false,
		      i);
    stuckPos += stuckPosShift;
  }

  // World PV
  G4ThreeVector worldPos(0, 0, 0);  
  G4VPhysicalVolume* world = new G4PVPlacement(0,
					       worldPos,
					       fLogicWorld,
					       "World",
					       0,
					       false,
					       0);

  // add SD
  if(!fScintillatorSD){
    G4SDManager* SDman = G4SDManager::GetSDMpointer();
    fScintillatorSD = new ScintillatorSD("/ScintillatorSD");
    SDman->AddNewDetector(fScintillatorSD);
  }
  fLogicNeut->SetSensitiveDetector(fScintillatorSD);
  if(fVetoNumber != 0) fLogicVeto->SetSensitiveDetector(fScintillatorSD);

  // colors
  fLogicWorld->SetVisAttributes(G4VisAttributes::Invisible);
  fLogicStuck->SetVisAttributes(G4VisAttributes::Invisible);
  fLogicNeutLayer->SetVisAttributes(G4VisAttributes::Invisible);
  if(fVetoNumber != 0) fLogicVetoLayer->SetVisAttributes(G4VisAttributes::Invisible);

  G4VisAttributes* neutCcolor = new G4VisAttributes(G4Colour::Yellow());
  fLogicNeut->SetVisAttributes(neutCcolor);
  G4VisAttributes* vetoCcolor = new G4VisAttributes(G4Colour::Cyan());
  if(fVetoNumber != 0) fLogicVeto->SetVisAttributes(vetoCcolor);
  
  // Print status
  G4cout << "### Neut consist of " 
         << fNeutMaterial->GetName() 
         <<  "  ###" << G4endl;

  G4cout <<* (G4Material::GetMaterialTable()) << G4endl;

  // save parameter which was used
  if(!gSimulationParameterNeutronDetector){
    gSimulationParameterNeutronDetector = new TSimulationParameterNeutronDetector("parameterNeutronDetector");
  }
  gSimulationParameterNeutronDetector->fNeutNumber = fNeutNumber;
  gSimulationParameterNeutronDetector->fNeutSize = 1/cm*TVector3(fNeutSize.x(),
								 fNeutSize.y(),
								 fNeutSize.z());
  gSimulationParameterNeutronDetector->fNeutGap = fNeutGap/cm;
  gSimulationParameterNeutronDetector->fNeutRotationMode = fNeutRotationMode;

  gSimulationParameterNeutronDetector->fVetoNumber = fVetoNumber;
  gSimulationParameterNeutronDetector->fVetoSize = 1/cm*TVector3(fVetoSize.x(),
								 fVetoSize.y(),
								 fVetoSize.z());
  gSimulationParameterNeutronDetector->fVetoGap = 1/cm*TVector3(fVetoGap.x(),
								fVetoGap.y(),
								fVetoGap.z());
  gSimulationParameterNeutronDetector->fVetoRotationMode = fVetoRotationMode;

  gSimulationParameterNeutronDetector->fNeutLayerNumber = fNeutLayerNumber;
  gSimulationParameterNeutronDetector->fNeutLayerGap = fNeutLayerGap/cm;

  gSimulationParameterNeutronDetector->fNeutVetoGap = fNeutVetoGap/cm;

  gSimulationParameterNeutronDetector->fStuckNumber = fStuckNumber;
  gSimulationParameterNeutronDetector->fStuckGap = fStuckGap/cm;

  gSimulationParameterNeutronDetector->fPosition = 1/cm*TVector3(fPosition.x(),
								 fPosition.y(),
								 fPosition.z());

  return world;
}

void DetectorConstruction::UpdateGeometry()
{
  G4RunManager::GetRunManager()->DefineWorldVolume(Construct());
}

void DetectorConstruction::SetNeutNumber(G4int val)  
{
  if(val > 0 && val != fNeutNumber){
    fNeutNumber = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutSizeX(G4double val)  
{
  if(val > 0 && val != fNeutSize.x()){
    fNeutSize.setX(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutSizeY(G4double val)  
{
  if(val > 0 && val != fNeutSize.y()){
    fNeutSize.setY(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutSizeZ(G4double val)  
{
  if(val > 0 && val != fNeutSize.z()){
    fNeutSize.setZ(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutGap(G4double val)  
{
  if(val >= 0 && val != fNeutGap){
    fNeutGap = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutRotationMode(G4int val)  
{
  if(val != fNeutRotationMode){
    fNeutRotationMode = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoNumber(G4int val)  
{
  if(val >= 0 && val != fVetoNumber){
    fVetoNumber = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoSizeX(G4double val)  
{
  if(val > 0 && val != fVetoSize.x()){
    fVetoSize.setX(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoSizeY(G4double val)  
{
  if(val > 0 && val != fVetoSize.y()){
    fVetoSize.setY(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoSizeZ(G4double val)  
{
  if(val > 0 && val != fVetoSize.z()){
    fVetoSize.setZ(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoGapY(G4double val)  
{
  if(val != fVetoGap.y()){
    fVetoGap.setY(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoGapZ(G4double val)  
{
  if(val != fVetoGap.z()){
    fVetoGap.setZ(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetVetoRotationMode(G4int val)  
{
  if(val != fVetoRotationMode){
    fVetoRotationMode = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutLayerNumber(G4int val)  
{
  if(val > 0 && val != fNeutLayerNumber){
    fNeutLayerNumber = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutLayerGap(G4double val)  
{
  if(val >= 0 && val != fNeutLayerGap){
    fNeutLayerGap = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetNeutVetoGap(G4double val)  
{
  if(val >= 0 && val != fNeutVetoGap){
    fNeutVetoGap = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetStuckNumber(G4int val)  
{
  if(val > 0 && val != fStuckNumber){
    fStuckNumber = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetStuckGap(G4double val)  
{
  if(val >= 0 && val != fStuckGap){
    fStuckGap = val;
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  } 
}

void DetectorConstruction::SetStuckInterval(G4double val)  
{
  G4double neutStuckSizeZ = (fNeutSize.z()+fNeutLayerGap)*fNeutLayerNumber-fNeutLayerGap;
  G4double vetoStuckSizeZ = fVetoSize.z()+fVetoGap.z()+fNeutVetoGap;

  G4double stuckGap = val - (neutStuckSizeZ + vetoStuckSizeZ);
  if(stuckGap < 0){
    G4cerr << "" << G4endl;
    G4cerr << " DetectorConstrucktion::SetStuckInterval> Stuck gap could not be less than 0 !!" << G4endl;
    G4cerr << " DetectorConstrucktion::SetStuckInterval> Set proper interval (or maybe better to set at the last)" << G4endl;
    G4cerr << " DetectorConstrucktion::SetStuckInterval> Current full width of stuck is "
	   << (neutStuckSizeZ + vetoStuckSizeZ)/cm << " cm"<< G4endl;
    G4cerr << "" << G4endl;
    G4Exception(0, 0, FatalException, 0);
  }else{
    SetStuckGap(stuckGap);
  }
}

void DetectorConstruction::SetPositionX(G4double val)  
{
  if(val != fPosition.x()){
    fPosition.setX(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  }
}

void DetectorConstruction::SetPositionY(G4double val)  
{
  if(val != fPosition.y()){
    fPosition.setY(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  }
}

void DetectorConstruction::SetPositionZ(G4double val)  
{
  if(val >= 0 && val != fPosition.z()){
    fPosition.setZ(val);
    G4RunManager::GetRunManager()->GeometryHasBeenModified();
  }
}

// void DetectorConstruction::SetNeutMaterial(const G4String& mat)
// {
//   G4Material* material = G4NistManager::Instance()->FindOrBuildMaterial(mat);
//   if(material && material != fNeutMaterial){
//     fNeutMaterial = material;
//     if(fLogicNeut) fLogicNeut->SetMaterial(fNeutMaterial);
//     G4RunManager::GetRunManager()->PhysicsHasBeenModified();
//   }
// }
// 
// void DetectorConstruction::SetWorldMaterial(const G4String& mat)
// {
//   G4Material* material = G4NistManager::Instance()->FindOrBuildMaterial(mat);
//   if(material && material != fWorldMaterial){
//     fWorldMaterial = material;
//     if(fLogicWorld) fLogicWorld->SetMaterial(fWorldMaterial);
//     G4RunManager::GetRunManager()->PhysicsHasBeenModified();
//   }
// }

